''nRF24LE1 Programming Demo
''
''
{{/////////////////////////////////////////////////////////////////////////////////////////////////
// ProperBurner_nRF - An SPI Flash writer for nRF24LE1 modules using the Propeller
//
// Author      : Alex Stanfield
// Updated     : 2/21/2014
// Designed For: P8X32A
// Version     : 0.2
//
// Copyright (c) 2014 Alex Stanfield
// See end of file for terms of use.
//
// Update History:
//
// v0.2 - Original release - 2/21/2014.
//
// Prereqs:
// SW (included)
// - SPI_Asm driver (Beau Schwabe)
// - Full Duplex Serial (Chip Gracey, Jeff Martin, Daniel Harris) 
// - HEX File to burn onto your module
// HW
// - nRF24LE1 module (possibly with a HW adapter for pin matching)
// - QuickStart or other Propeller based board 
//
///////////////////////////////////////////////////////////////////////////////////////////////////
// General rules for usage:
//
// Disclaimer: If you continue you will be connecting two active devices,
//             follow the "check twice, build once" rule
//
// For troubleshooting use this guide:
//
// case RESULT_OF_YOUR_EXPERIMENT
//   OK           :  MY_FAULT_AND_YOU_DID_A_GOOD_JOB___ENJOY_IT
//   SMOKE,
//   EXPLOSION,
//   BAD_THINGS,
//   NOTHING      : YOUR_FAULT__CHECK_EVERYTHING_AGAIN
//   other        : YOUR_FAULT__CHECK_EVERYTHING_AGAIN
//        
///////////////////////////////////////////////////////////////////////////////////////////////////
// Setting up the Flash Burner
// 
// 1 - You need to hook up the following pins of your nRF24LE1 module to the Propeller
//     (see the CON section)

  Power pins for nRF24LE1 module. check your module for other stuff that might use power, mine had
  only the nRF24LE1 chip which you can power from the Propeller pins. If you also plan to RUN the module
  powered from the Propeller pins do the math with all the functions you will enable with your program

  I believe you can run the whole thing powered from the propeller but do that at your own risk! 

  nRF_PROG      'O |Set to 1 to program module (15)
  nRF_RESET     'O |Set to 0 to reset module  (27)

  ''SPI interface. The Propeller acts as the MASTER, nRF24LE1 as SLAVE
  nRF_CS        'Output  SPI  Chip Select
  nRF_MISO      'Input   SPI  Master Input  Slave Output
  nRF_MOSI      'Output  SPI  Master Output Slave Input
  nRF_SCK       'Output  SPI  Serial Clock

// 2 - Configure the CON section to match the pins used on your propeller
//
// 3 - Configure the filename of you HEX file at the end of this program (instruction file"blink.ihx")
//
// 4 - Load & run this program and watch the terminal do the checkings and verification
//
// You can change the behavior of the program using the constants AUTO_MODE_...  (see below) 
//
///////////////////////////////////////////////////////////////////////////////////////////////////
}}

CON
  _xinfreq  = 5_000_000        
  _clkmode  = xtal1 + pll16x

  bp = 8        'base pin for nRF24LE1 module

  nRF_VDD       = bp         'O |Power pins : 3V3  (-1=NOT powered from a propeller pin)
  nRF_GND       = bp + 1     'O |Power pins : Gnd  (-1=NOT powered from a propeller pin)

  nRF_P15       = bp + 2     'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx not used in demo
  nRF_P16       = bp + 3     'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "
  nRF_P00       = bp + 4     'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "
  nRF_P01       = bp + 5     'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "
  nRF_P02       = bp + 6     'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "

  nRF_PROG      = bp + 7     'O |Set to 1 to program module (15)

  nRF_P03       = bp + 8     'I |leds in QS  (16)
  nRF_P04       = bp + 9     'I ||
  '-------------- bp + 10 <<< not used, wire skipped------------
  nRF_P05       = bp + 11    'O |SPI_SCK  (QS:led pin 19)
  nRF_P06       = bp + 12    'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx not used in demo
  nRF_P07       = bp + 13    'O |SPI_MOSI (QS:led pin 21)
  nRF_P10       = bp + 14    'I |SPI_MISO (QS:led pin 22)
  nRF_P11       = bp + 15    'O |SPI_CS (led on pin 23 from QS)
  nRF_P12       = bp + 16    'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx not used in demo
  nRF_P13       = bp + 17    'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "
  nRF_P14       = bp + 18    'I |xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "

  nRF_RESET     = bp + 19    'O |Set to 0 to reset module  (27)

  nRF_CS        = bp + 15       'Output
  nRF_MISO      = bp + 14       'Input
  nRF_MOSI      = bp + 13       'Output from Prop to module
  nRF_SCK       = bp + 11       'Output

  'Definitions--------------------------------------------------------------
  YES           = true
  NO            = false
  'Default behavior---------------------------------------------------------
  AUTO_MODE       = YES         'Auto or Interactive? (use YES, interactive is not there yet)
  AUTO_MODE_CHECK = NO          'Check HEX file integrity (forced when PROG/VERIFY is done)
  AUTO_MODE_PROG  = YES         'Should we program the module?
  AUTO_MODE_VERIFY= YES         'Should we verify program vs file?
  AUTO_MODE_RUN   = YES         'Should we let the module run after?            

  '---Hex Parsing type------------------------------------------------------
  #0, CHECK_FILE, WRITE, VERIFY 'Parameter for Parse_Hex routine

  
  '---Serial Terminal config------------------------------------------------
  rxPin  = 31                   'serial
  txPin  = 30
  baud   = 115200

  'SPI commands for nRF24LE1------------------------------------------------
                        'Add|Data|Description
                        '---+----+-------------------------------------------    
  nRF_CMD_WREN   = $06  'n/a|  0 |Enable  write to Flash (set WEN bit in FSR)
  nRF_CMD_WRDIS  = $04  'n/a|  0 |Disable write to flash

  nRF_CMD_RDSR   = $05  'n/a|  1 |Read  Flash Status Reg (FSR)
  nRF_CMD_WRSR   = $01  'n/a|  1 |Write Flash Status Reg (FSR)

  nRF_CMD_READ   = $03  ' 2 |>=1 |Read  Data from Flash
  nRF_CMD_WRITE  = $02  ' 2 |>=1 |Write Data to Flash

  nRF_CMD_RDFPCR = $89  'n/a|  1 |Read Flash Config Register

  nRF_CMD_ERASE  = $52  ' 1 |  0 |Erase addressed page)
  nRF_CMD_ERALL  = $62  'n/a|  0 |See manual before using  DANGER!!

  nRF_CMD_RDISMB = $85  ' 1 |  0 |Set Flash readback protection DANGER!!
  nRF_CMD_DEBUG  = $86  'n/a|  0 |Enable HW debug features

  
obj
  fdx   :       "fullduplexserial"
  spi   :       "SPI_asm"

VAR
  long  rd_byte1, rd_byte2, rd_byte3, rd_byte4
  long  _1us, _1ms, _1s         'set to 1us/1ms/1s

  'nRF24Le1 Data Structure--------------------------------------------
  long nRF_var_FSR              'FSR  -  Flash Status Register
  long nRF_var_PCR              'PCT  -  Protect Control Register                                      

pub main | z,lst_z
  _1us := clkfreq / 1_000_000                            ' calc 1us for current xtal
  _1ms := clkfreq / 1_000
  _1s  := clkfreq

  'Set default values for lines (before driving outputs for real)-------------
  outa[nRF_RESET]  := 1    'non-reset state
  outa[nRF_PROG]   := 0    'running mode

  if nRF_VDD <> -1
    outa[nRF_VDD]  := 1    'Power on module
    dira[nRF_VDD]  := 1    '|

  if nRF_GND <> -1
    outa[nRF_GND]  := 0    '|
    dira[nRF_GND]  := 1    '|
    
  'Set outputs-----------------------------------------------------------------
  dira[nRF_RESET]  := 1    'These are always outputs
  dira[nRF_PROG]   := 1    '|

  'Set console terminal--------------------------------------------------------
  fdx.start(rxPin, txPin, 0, baud)                      ' init serial port
  waitcnt(_1s * 3 + cnt)
  fdx.tx(0)                                             ' cls
  fdx.str(string(13,"nRF24LE1 Programming Demo",13,13))

  if AUTO_MODE

    if AUTO_MODE_CHECK OR AUTO_MODE_PROG OR AUTO_MODE_VERIFY
      if \Parse_HEX_File(CHECK_FILE)
        'shutdown
        fdx.str(string(13,"should abort"))

    if AUTO_MODE_PROG
      Reboot_nRF(true{PROG})    'Set module in PROG mode
      Erase_All                 '
      if \Parse_HEX_File(WRITE) 'Write program to flash

    if AUTO_MODE_VERIFY
      Reboot_nRF(true{PROG})     'Reboot to verify
      if \Parse_HEX_FIle(VERIFY)    'Verify programm correctly written
        'shutdown

{   Dump_Flash
    fdx.tx(13)
    fdx.bin(Cmd_RDSR(@nRF_var_FSR),8)
}
    if AUTO_MODE_RUN
      Reboot_nRF(false{RUN})     'Reboot to run your own code, keep module powered

      '/// following code is to track bit changes with the module running. Remove or adapt
      '/// to your needs
      lst_z := true
      repeat
        z := ina[nRF_P07..nRF_P03]
        if z <> lst_z
          fdx.tx(13)
          fdx.bin(z,8)
          lst_z := z
    
  else

''end program here---------
  SPI.stop
  
pri ShutDown
  'End all cogs------
  spi.stop
  fdx.stop
  cogstop(cogid)

'////////////////////////////////////////////////////////////////////////////////////////
'///  SPI ROUTINES                                                                    ///
'////////////////////////////////////////////////////////////////////////////////////////

pub SPI_cmd0(par_cmd)

  CS_on
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_cmd)       'Send command to module
  SPI_end

pub SPI_cmd1(par_cmd, par_add)

  CS_on
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_cmd)       'Send command to module
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_add)       'Send add/value
  SPI_end

pub SPI_cmd2(par_cmd, par_add)

  CS_on
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_cmd)       'Send command to module
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_add >>   8)'Send Address High byte                 
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_add &  $FF)'Send Address Low  byte

pub SPI_Read1(par_cmd, ptr_var)
''---------------------------------------------------------------------------------------
'' Send cmd and read back 1 byte from SPI device
''---------------------------------------------------------------------------------------

  CS_on
  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_cmd)                     'Send command to module 
  result := SPI_Read(ptr_var)

pub SPI_Read(ptr_var)
''---------------------------------------------------------------------------------------
'' Read 1 byte from SPI device (init transaction with Cmd_READ_Init)
''---------------------------------------------------------------------------------------

  result := spi.shiftin(nRF_MISO, nRF_SCK, spi#MSBPRE, 8)
  long[ptr_var] := result

pub SPI_Write(par_value)
''---------------------------------------------------------------------------------------
'' Write 1 byte to SPI device (init transaction with Cmd_WRITE_Init)
''---------------------------------------------------------------------------------------

  spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 8, par_value)     'Send add/value

pri CS_on                       'Select chip---------------------------
  outa[nRF_CS] := 0

pub SPI_end
''---------------------------------------------------------------------------------------
'' Finish transaction with device
''---------------------------------------------------------------------------------------

  delay(_1us*25)                'Wait enough for last byte being written (if any)
  outa[nRF_CS] := 1             'De-select chip

'////////////////////////////////////////////////////////////////////////////////////////
'///  nRF24LE1 specific commands                                                      ///
'////////////////////////////////////////////////////////////////////////////////////////
pub Cmd_WREN
''---------------------------------------------------------------------------------------
''  Set WRITE ENABLE in module
''---------------------------------------------------------------------------------------

' fdx.str(string(13,"(Dbg -> WREN)",13))
  Wait_Flash
  SPI_cmd0(nRF_CMD_WREN)

pub Cmd_WRDIS
''---------------------------------------------------------------------------------------
''  Disable WRITE to flash/nvram in module
''---------------------------------------------------------------------------------------
  
  Wait_Flash
  SPI_cmd0(nRF_CMD_WRDIS)

pub Cmd_RDSR(ptr_var)
''---------------------------------------------------------------------------------------
''  Read Flash Status Register
''---------------------------------------------------------------------------------------
  result := SPI_read1(nRF_CMD_RDSR, ptr_var)
' fdx.tx("{")
' fdx.hex(result,2)
' fdx.tx("}")
  SPI_end

pub Cmd_WRSR(par_FSR)
''---------------------------------------------------------------------------------------
''  Write back Status Register
''---------------------------------------------------------------------------------------
  Wait_Flash
  SPI_cmd1(nRF_CMD_WRSR, par_FSR)

pub Cmd_RDFPCR(ptr_var)
''---------------------------------------------------------------------------------------
''  Read Flash Protection Configuration Register
''---------------------------------------------------------------------------------------

  Wait_Flash
  result := SPI_Read1(nRF_CMD_RDFPCR, ptr_var)
  SPI_End

pub Cmd_READ_Init(par_add)
''---------------------------------------------------------------------------------------
''  Initialize READ operation
''---------------------------------------------------------------------------------------

  Wait_Flash
  SPI_cmd2(nRF_CMD_READ, par_add)
  
pub Cmd_WRITE_Init(par_add)
''---------------------------------------------------------------------------------------
''  Initialize WRITE operation  (also set WEN bit)
''---------------------------------------------------------------------------------------

  Wait_Flash
  Cmd_WREN
  SPI_cmd2(nRF_CMD_WRITE, par_add)

pub Wait_Flash | z

  if Cmd_RDSR(@z) & %0001_0000
    fdx.str(string(13,"Waiting for Flash to become ready"))
    repeat while Cmd_RDSR(@z) & %0001_0000
      fdx.tx(".")
      delay(_1us*10)
    fdx.str(string("[Ready]"))

 
'////////////////////////////////////////////////////////////////////////////////////////
PRI delay(time)
  waitcnt(time + cnt)                                   ' timed delay    

DAT 'InfoPage data
IPAGE
DSYS    byte $00[32]
NUPP    byte $00
RES1    byte $AA[2]
RDISMB  byte $00
ENDEBUG byte $00
RES2    byte $AA[219]
USER    byte $00[256]
IPAGE_END

PRI Read_InfoPage | ip,z

  fdx.str(string(13,"-----------------------------------------------------------------"))
  fdx.str(string(13,"Saving InfoPage data in Propeller's hubram..."))
  '---Set pointer to INFOPAGE--------------------------------------------------------
  INFEN_On

  '---Read InfoPage------------------------------------------------------------------
  Cmd_READ_Init(0)
  repeat ip from 0 to 511
    IPAGE[ip] := spi.shiftin(nRF_MISO, nRF_SCK, spi#MSBPRE, 8)
  SPI_end

  '---Print InfoPage by sections------------------------------------------------------
  fdx.str(string(13,"InfoPage data   #### SAVE THIS SECTION ####"))
  fdx.str(string(13,"DSYS   :"))
  repeat ip from 0 to 31
    fdx.hex(DSYS[ip],2)
    fdx.tx(" ")

  fdx.str(string(13,"NUPP   :"))
  fdx.dec(NUPP)
  fdx.str(string(13,"RDISMB :"))
  fdx.hex(RDISMB,2)
  fdx.str(string(13,"ENDEBUG:"))
  fdx.hex(ENDEBUG,2)
    
  fdx.str(string(13,"USER   :"))
  ip~
  repeat while ip < 256
    repeat 16
      fdx.hex(USER[ip++],2)
      fdx.tx(" ")
    fdx.str(string(13,"        "))

  fdx.str(string(13,"#### END OF INFOPAGE ######################"))

  INFEN_Off

PRI Write_InfoPage | ip, z

  fdx.str(string(13,"-----------------------------------------------------------------"))
  fdx.str(string(13,"Writing back InfoPage data to nRF24LE1..."))

  INFEN_On        'enable InfoPage Write

  '---write back DSYS and rest of InfoPage-----------------------
  fdx.str(string(13,"..writing"))
  Cmd_WRITE_Init($000) '($000) 'XXXXXXXX
  repeat ip from 0 to 511
    SPI_Write(IPAGE[ip])
  SPI_end
  fdx.str(string(13,"..ended"))
  Wait_Flash
  
  INFEN_Off

PRI INFEN_On
  fdx.str(string(13,"Setting InfoPage (INFEN) bit in FSR: "))
  Wait_Flash
  fdx.bin(Cmd_RDSR(@nRF_var_FSR),8)
  
  '---Set INFEN in FSR--------------------------------------------------------------
  nRF_var_FSR |= %1000   'InfoPage enable bit
  Cmd_WRSR(nRF_var_FSR)
  
  fdx.str(string(" -> "))
  fdx.bin(Cmd_RDSR(@nRF_var_FSR),8)

PRI INFEN_Off
  fdx.str(string(13,"Clearing InfoPage (INFEN) bit in FSR: "))
  Wait_Flash
  fdx.bin(Cmd_RDSR(@nRF_var_FSR),8)
  
  '---Set INFEN in FSR--------------------------------------------------------------
  nRF_var_FSR &= %11110111 'Clear InfoPage enable bit
  Cmd_WRSR(nRF_var_FSR)
  
  fdx.str(string(" -> "))
  fdx.bin(Cmd_RDSR(@nRF_var_FSR),8)


PRI Dump_Flash | ip,z

  INFEN_Off

  '---Read InfoPage------------------------------------------------------------------
  fdx.str(string(13,"Dumping flash contents"))
  Cmd_READ_Init(0)
  ip~
  repeat 8
    repeat 32
      fdx.tx(13)
      fdx.hex(ip,4)
      fdx.tx(":")
      repeat 32
        fdx.hex(SPI_Read(@z),2)
        fdx.tx(" ")
        ip++
  SPI_end

PRI Read_NVRAM

PRI Write_NVRAM

PUB Erase_All
''---------------------------------------------------------------
''Erase_All                     Clear all flash AND InfoPage data
''
''This routine first saves the InfoPAge data for later restore
''after erasing all flash data
''---------------------------------------------------------------

  Read_InfoPage                 'Save to InfoPage to hubram

  '---Proceed to erase flash ram --------------------------------
  fdx.str(string(13,"-----------------------------------------------------------------"))
  fdx.str(string(13,"Erasing flash"))
  'send erase_all cmd
  INFEN_Off
  Wait_Flash
  Cmd_WREN
  SPI_Cmd0(nRF_CMD_ERALL)
  'repeat until flash ready (FSYN == 0) in FSR
  Wait_Flash

  Write_InfoPage                'Write back InfoPage

PRI Erase_Page(par_page)

PRI Reboot_nRF(par_Prog)
''---------------------------------------------------------------
''Reboot_nRF(par_Prog)          Reboot nRF24LE1 module
''
''par_Prog                      false = run normally
''                              true  = set module in PROG mode
''---------------------------------------------------------------

  fdx.str(string(13,"rebooting with "))
  if par_prog
    fdx.str(string("PROG"))
  else
    fdx.str(string("RUN"))

  outa[nRF_PROG]:=par_Prog & %1 'Set Prog mode

  'Pulse RESET-----------------------------------------------------
  outa[nRF_RESET]  := 0
  delay(_1us<<3) 'us*20)          'Ensure a good reset pulse

  '--- Set lines according to RUN/PROG request---------------------
  if par_Prog
    outa[nRF_CS]     := 1       'Chip NOT selected
    '--- Start SPI interface---------------------------------------
    dira[nRF_CS]     := 1       'CS line is output
    spi.Start(15, 0)            '15=1uS delay between line changes;  0=clk low
    spi.shiftout(nRF_MOSI, nRF_SCK, spi#MSBFIRST, 1, 0)  'Done to init default state for MOSI
  else
    spi.Stop                    'Stop SPI interface
    dira[nRF_CS]     := 0       'CS is now an input to allow module's P11

  outa[nRF_RESET]  := 1         'Boot up
  delay(_1mS * 2)               'Wait for flash

  fdx.str(string(", finished reboot....<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"))


var
  long  ix, records, c_checksum, dx, TotData
  long  h_len, h_add, h_type, h_data, h_checksum
  byte  pages[36]

PRI Parse_Hex_File(par_type) | z
''Hex file parser-------------------------------------

fdx.str(string(13,"-----------------------------------------------------------------"))
case par_type
  CHECK_FILE: fdx.str(string(13,"Verifying HEX file..."))
  WRITE     : fdx.str(string(13,"WRITING flash..."))
              INFEN_Off
  VERIFY    : fdx.str(string(13,"VERIFYING flash..."))
              INFEN_Off
  other     : fdx.str(string(13,"Invalid type, aborting 'Parse_Hex_File'"))
              abort 2
               
records~
TotData~
repeat ix from @hex to @tail
  if byte[ix] == ":"
    records++
    h_len := hh2dec(ix+1)
    h_add := hh2dec(ix+3)* 256 + hh2dec(ix+5)
    h_type:= hh2dec(ix+7)
    c_checksum := h_len + (h_add & $FF) + (h_add>>8) + h_type
    pages[(h_add>>9)] := 1

    fdx.str(string(13,"Rec: "))
    if records < 100
      fdx.tx(" ")
    if records < 10
      fdx.tx(" ")
    fdx.dec(records)
    fdx.str(string("|"))
    fdx.hex(h_len,2)
    fdx.tx(" ")
    fdx.hex(h_add,4)
    fdx.tx(" ")
    fdx.hex(h_type,2)
    ifnot h_type < 2
      fdx.tx("*")
      result := 1
    fdx.str(string(" > "))
    ix += 9

    if h_type == 0
      case par_type
        CHECK_FILE:
        WRITE     : Cmd_WRITE_Init(h_add) 'send write cmd with address
        VERIFY    : Cmd_READ_Init(h_add)  'send read  cmd with address
        other: abort
        
    dx~                         'Reset byte counter for record
    repeat while dx < h_len     'Process all bytes in hex record
      h_data := hh2dec(ix+dx<<1) 
      c_checksum += h_data
      fdx.hex(h_data,2)

      if h_type == 0
        case par_type
          CHECK_FILE:
          WRITE     : SPI_Write(h_data) 'send data to device
          VERIFY    : 'verify byte on device is the same as in file
            if h_data <> SPI_read(@z)
              fdx.tx("/")
              fdx.hex(z,2)
              result := 1         'Flag error during verify
          other: abort
         
      fdx.tx(" ")
      dx++
    fdx.tx("<")                 'signal payload end

    SPI_End                     'Force SPI transaction end

    ix += dx<<1                 'adjust byte pointer in file
    h_checksum := hh2dec(ix)
    c_checksum := (-c_checksum) & $FF
    fdx.hex(h_checksum,2)
    
    if h_checksum <> c_checksum
      fdx.tx(" ")
      fdx.hex(c_checksum,2)
        fdx.str(string("<<< checksum ERROR, aborting"))
      abort

    TotData += h_len
    
    ix + = 2                    'Point to CR byte & search for next record

if par_type == CHECK_FILE
  if h_type == 1 AND result == 0
    fdx.str(string(13,"HEX file is OK"))
  else
    fdx.str(string(13,"Errors found in HEX file"))
   
  fdx.str(string(13, "Total data [bytes]: "))
  fdx.dec(TotData)
   
  fdx.str(string(13,"Pages used",13))
  repeat ix from 0 to 35
    fdx.hex(pages[ix],1)
    fdx.tx(" ")
   
elseif result
  fdx.str(string(13,"#### ERRORS FOUND DURING WRITE/VERIFY ####"))
  abort result

case par_type
  CHECK_FILE: fdx.str(string(13,"CHECK FILE"))
  WRITE     : fdx.str(string(13,"WRITE"))
  VERIFY    : fdx.str(string(13,"VERIFY"))
fdx.str(string(" operation finished OK!!!"))
return 0    
   

pri hh2dec(par_add) | t

  if (     t := h2d(par_add)) < 0
    return t
  if (result := h2d(par_add+1)) < 0
    return
  result += t * 16 

pri h2d(par_add) | t

 case t:=byte[par_add]
    "0".."9":result := t - "0"
    "a".."f","A".."F": result := (t & $0F) + 9
    other: result := -1

DAT
  hex  file "blink.ihx"         'Place your HEX format file here
  tail                          'Label used to know EOF from embedded HEX file